package com.privas.application.dto.mapper;

import com.privas.application.dto.*;
import com.privas.application.services.*;
import com.privas.application.services.RepositoryType;
import com.privas.application.dto.*;
import com.privas.application.services.*;

import java.util.Collection;
import java.util.List;

import static java.util.stream.Collectors.toList;

public class PrivasLMapper {

    public static PrivasLData map(PrivacyDefinition privacyDefinition) {
        PrivasLData privasLData = new PrivasLData();

        privasLData.setRepository(mapRepository(privacyDefinition));
        privasLData.setPrivacyDefinition(mapPrivacyTechniquesDefinition(privacyDefinition.getAttackModels()));
        privasLData.setEntities(mapEntities(privacyDefinition.getTableToPrivateDetails()));
        privasLData.setGeneralizationTrees(mapTrees(privacyDefinition.getGeneralizationsUnique()));

        return privasLData;
    }

    private static Repository mapRepository(PrivacyDefinition privacyDefinition) {
        Repository repository = new Repository();
        MySQLRepository mySQLRepository = new MySQLRepository();
        CsvRepository csvRepository = new CsvRepository();

        if (privacyDefinition.getRepositoryType().equals(RepositoryType.MYSQL)) {
            ConnectionStringJDBC connectionString = ConnectionStringJDBC.getFromString(privacyDefinition.getConnectionString());

            mySQLRepository.setUrl(connectionString.getUrl());
            mySQLRepository.setPort(connectionString.getPort());
            mySQLRepository.setDatabaseName(connectionString.getDatabaseName());
            mySQLRepository.setUsername(connectionString.getUsername());
            mySQLRepository.setPassword(connectionString.getPassword());
            repository.setSelectedKind(com.privas.application.dto.RepositoryType.MYSQL);
        } else {
            csvRepository.setPath(privacyDefinition.getPath());
            repository.setSelectedKind(com.privas.application.dto.RepositoryType.CSV);
        }
        repository.setMysqlData(mySQLRepository);
        repository.setCsvData(csvRepository);

        return repository;
    }

    private static PrivacyTechniquesDefinition mapPrivacyTechniquesDefinition(List<AttackModel> attackModels) {
        PrivacyTechniquesDefinition techniquesDefinition = new PrivacyTechniquesDefinition();

        techniquesDefinition.setRecordLinkage(attackModels.contains(AttackModel.RECORD_LINKAGE));
        techniquesDefinition.setAttributeLinkage(attackModels.contains(AttackModel.ATTRIBUTE_LINKAGE));
        techniquesDefinition.setProbabilisticAttack(attackModels.contains(AttackModel.PROBABILISTIC_ATTACK));
        techniquesDefinition.setTableLinkage(attackModels.contains(AttackModel.TABLE_LINKAGE));

        return techniquesDefinition;
    }

    private static List<Entity> mapEntities(List<TableToPrivateDetails> tables) {
        return tables.stream().map(PrivasLMapper::mapEntity).collect(toList());
    }

    private static Entity mapEntity(TableToPrivateDetails table) {
        Entity entity = new Entity();
        entity.setTitle(table.getTableName());

        List<EntityAttribute> entityAttributes = table.getAttributesOptions().getAttributes().stream()
                .map(PrivasLMapper::mapEntityAttribute)
                .collect(toList());

        entity.setAttributes(entityAttributes);

        return entity;
    }

    private static EntityAttribute mapEntityAttribute(Attribute attribute) {
        EntityAttribute entityAttribute = new EntityAttribute();

        entityAttribute.setName(attribute.getName());

        if (attribute.hasType()) {
            entityAttribute.setTaxonomy(attribute.getType());
        }

        if (attribute.isId()) {
            entityAttribute.setAttributeType(PrivacyAttributeType.ID);
        } else if (attribute.isQId()) {
            entityAttribute.setAttributeType(PrivacyAttributeType.QID);
        } else if (attribute.isSA()) {
            entityAttribute.setAttributeType(PrivacyAttributeType.SA);
        }

        return entityAttribute;
    }

    private static List<Tree> mapTrees(Collection<Generalization> generalizations) {
        return generalizations.stream().map(PrivasLMapper::mapTree).collect(toList());
    }

    private static Tree mapTree(Generalization generalization) {
        Tree tree = new Tree();
        tree.setTitle(generalization.getName());

        List<TreeData> data = generalization.getGeneralizationNodes().stream()
                .map(PrivasLMapper::mapTreeData)
                .collect(toList());

        tree.setData(data);

        return tree;
    }

    private static TreeData mapTreeData(GeneralizationNode node) {
        TreeData treeData = new TreeData();
        treeData.setLabel(node.getLabel());

        if (node instanceof GeneralizationNodeRange || node instanceof GeneralizationNodeSetValues) {
            treeData.setValue(node.toString());
        } else {
            List<TreeData> children = node.getChildren().stream().map(PrivasLMapper::mapTreeData).collect(toList());
            treeData.setChildren(children);
        }

        return treeData;
    }
}
