package com.privas.application.services;

import com.privas.application.dto.*;
import com.privas.application.dto.*;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

public class PrivasLGenerator {

    public static String generate(PrivasLData privasLData) {
        String repository = generateRepository(privasLData.getRepository());
        String prevention = generatePrevention(privasLData.getPrivacyDefinition());
        String entities = generateEntities(privasLData.getRepository(), privasLData.getEntities());
        String generalizationTrees = generateGeneralizationTrees(privasLData.getGeneralizationTrees());

        return String.format("%s\n\n%s\n\n%s\n\n%s", repository, entities, prevention, generalizationTrees);
    }

    private static String generateRepository(Repository repository) {
        String repositoryOutput = "";

        switch (repository.getSelectedKind()) {
            case MYSQL:
                repositoryOutput = generateMySQLRepository(repository.getMysqlData());
                break;
            case CSV:
                repositoryOutput = generateCsvRepository(repository.getCsvData());
                break;
        }

        return String.format("%s", repositoryOutput);
    }

    private static String generateMySQLRepository(MySQLRepository mySQLRepository){
        return String.format("Relational\n  Connection String: \"%s\"", generateConnectionString(mySQLRepository));
    }

    private static String generateConnectionString(MySQLRepository mySQLRepository) {
        StringBuilder stringBuilder = new StringBuilder(String.format("jdbc:mysql://%s:%s/%s?", mySQLRepository.getUrl(), mySQLRepository.getPort(), mySQLRepository.getDatabaseName()));

        List<String> attributes = new ArrayList<>();

        if (mySQLRepository.getUsername() != null && !mySQLRepository.getUsername().isEmpty()) {

            attributes.add(String.format("user=%s", mySQLRepository.getUsername()));
        }

        if (mySQLRepository.getPassword() != null && !mySQLRepository.getPassword().isEmpty()) {
            attributes.add(String.format("password=%s", mySQLRepository.getPassword()));
        }

        stringBuilder.append(String.join("&", attributes));

        return stringBuilder.toString();
    }

    private static String generateCsvRepository(CsvRepository csvRepository) {
        return String.format("CSV\n  Path: \"%s\"", csvRepository.getPath());
    }

    private static String generatePrevention(PrivacyTechniquesDefinition privacyDefinition) {
        List<String> preventions = generatePreventions(privacyDefinition);
        return String.format("Prevent From: [ %s ]", String.join(", ", preventions));
    }

    private static List<String> generatePreventions(PrivacyTechniquesDefinition privacyDefinition) {
        List<String> preventions = new ArrayList<>();

        if (privacyDefinition.isRecordLinkage()) {
            preventions.add("recordLinkage");
        }

        if (privacyDefinition.isAttributeLinkage()) {
            preventions.add("attributeLinkage");
        }

        if (privacyDefinition.isTableLinkage()) {
            preventions.add("tableLinkage");
        }

        if (privacyDefinition.isProbabilisticAttack()) {
            preventions.add("probabilisticAttack");
        }

        return preventions;
    }

    private static String generateEntities(Repository repository, List<Entity> entities) {
        String entitiesOutput = entities.stream()
                                        .map(PrivasLGenerator::generateEntity)
                                        .collect(Collectors.joining("\n  "));

        return String.format("%s\n  %s", repository.getRepositoryName(), entitiesOutput);
    }

    private static String generateEntity(Entity entity) {
        String attributes = entity.getAttributes()
                                  .stream()
                                  .map(PrivasLGenerator::generateAttribute)
                                  .collect(Collectors.joining(", "));

        return String.format("%s [ %s ]", entity.getTitle(), attributes);
    }

    private static String generateAttribute(EntityAttribute entityAttribute) {
        String type = generateAttributeType(entityAttribute.getAttributeType());
        String generalization = generateAttributeGeneralization(entityAttribute);

        return String.format("%s%s%s", type, entityAttribute.getName(), generalization);
    }

    private static String generateAttributeType(PrivacyAttributeType privacyAttributeType) {
        if (privacyAttributeType == null) {
            return "";
        }

        switch (privacyAttributeType) {
            case ID:
                return "@";
            case QID:
                return "&";
            case SA:
                return "~";
            default:
                return "";
        }
    }

    private static String generateAttributeGeneralization(EntityAttribute entityAttribute) {
        if (entityAttribute.hasGeneralization()) {
            return "::" + entityAttribute.getTaxonomy();
        }

        return "";
    }

    private static String generateGeneralizationTrees(List<Tree> generalizationTrees) {
        return generalizationTrees.stream()
                                  .map(PrivasLGenerator::generateGeneralizationTree)
                                  .collect(Collectors.joining("\n"));
    }

    private static String generateGeneralizationTree(Tree generalizationTree) {
        String children = generateGeneralizationSubTrees(generalizationTree.getData(), "  ");

        return String.format("+ %s\n%s", generalizationTree.getTitle(), children);
    }

    private static String generateGeneralizationSubTrees(List<TreeData> generalizationSubTrees, String prefix) {
        StringBuilder stringBuilder = new StringBuilder();

        for (TreeData subTree: generalizationSubTrees) {
            if (subTree.isLeaf()) {
                stringBuilder.append(generateGeneralizationLeafTree(subTree, prefix));
            }
            else {
                stringBuilder.append(generateGeneralizationSubTree(subTree, prefix));
            }
        }

        return stringBuilder.toString();
    }

    private static String generateGeneralizationLeafTree(TreeData generalizationSubTree, String prefix) {
        String label = getGeneralizationLabel(generalizationSubTree.getLabel());
        String value = getGeneralizationValue(generalizationSubTree.getValue());

        return String.format("%s> %s: %s\n", prefix, label, value);
    }

    private static String generateGeneralizationSubTree(TreeData generalizationSubTree, String prefix) {
        String label = getGeneralizationLabel(generalizationSubTree.getLabel());
        String nextLevelPrefix = prefix.concat("  ");

        String childrenOutput = generateGeneralizationSubTrees(generalizationSubTree.getChildren(), nextLevelPrefix);

        return String.format("%s+ %s\n%s", prefix, label, childrenOutput);
    }

    private static String getGeneralizationLabel(String labelRaw) {
        if (getNumberOfSpaces(labelRaw) == 0) {
            return labelRaw;
        }

        return String.format("'%s'", labelRaw);
    }

    private static String getGeneralizationValue(String value) {
        if (value.startsWith("[")) {
            return value;
        }

        return String.format("{ %s }", value);
    }
    
    private static int getNumberOfSpaces(String str) {
        int count = 0;

        if (str != null) {
            for (int i = 0; i < str.length(); i++) {
                if (Character.isWhitespace(str.charAt(i))) {
                    count++;
                }
            }
        }

        return count;
    }
}
